/*
 * candidate-p.c - Determine if a block device is a valid candidate.
 *
 * Oracle wants to only use partitions, not whole disks.  
 * That way, fdisk(8) won't hurt anything permanent.  However, certain
 * block devices have no partitions.  loop, dm, md, etc.  So, this
 * program has a simple semantic - if the device is considered a
 * valid candidate, it returns 0.
 *
 * Copyright (c) 2002-2006 Oracle Corporation.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE     /* For open64() */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

#include <linux/types.h>

#include "devmap.h"
#include "usage.h"

/* Kernel ABI ioctls for testing disks */
#define HDIO_GETGEO             0x0301

struct hd_geometry {
    unsigned char heads;
    unsigned char sectors;
    unsigned short cylinders;
    unsigned long start;
};



/*
 * Globals
 */
#define PROGNAME "oracleasm-candidate-p"
const char *program_name = PROGNAME;
const char *program_usage = (
    "Usage: " PROGNAME " [-f] <device>\n"
    "       " PROGNAME " -h\n"
    "       " PROGNAME " -V\n"
);
static int verbose;



static int device_is_partition(int fd)
{
    struct hd_geometry geo;
    int rc;

    rc = ioctl(fd, HDIO_GETGEO, &geo);
    if (rc)
        return 0;

    return !!geo.start;
}  /* device_is_partition() */


static int device_needs_partition(int fd)
{
    if (device_is_block(fd, "loop"))
        return 0;

    if (device_is_block(fd, "md"))
        return 0;

    if (device_is_block(fd, "device-mapper"))
        return 0;

    return 1;
}


static int open_disk(const char *disk_name)
{
    int fd, rc;
    struct stat64 stat_buf;

    rc = stat64(disk_name, &stat_buf);
    if (rc)
    {
        rc = -errno;
        if (verbose)
            error("Unable to query file \"%s\": %s\n", disk_name,
                  strerror(-rc));
        goto out;
    }

    /* First check, try to avoid opening tapes */
    if (!S_ISBLK(stat_buf.st_mode))
    {
        rc = -ENOTBLK;
        if (verbose)
            error("File \"%s\" is not a block device\n", disk_name);
        goto out;
    }

    fd = open64(disk_name, O_RDWR | O_EXCL);
    if (fd < 0)
    {
        rc = -errno;
        if (verbose)
            error("Unable to open device \"%s\": %s\n", disk_name,
                  strerror(-rc));
        goto out;
    }

    /* Second check, try not to write the wrong place */
    rc = fstat64(fd, &stat_buf);
    if (rc)
    {
        rc = -errno;
        close(fd);
        if (verbose)
            error("Unable to query device \"%s\": %s\n", disk_name,
                  strerror(-rc));
        goto out;
    }
    
    if (!S_ISBLK(stat_buf.st_mode))
    {
        rc = -ENOTBLK;
        close(fd);
        if (verbose)
            error("Device \"%s\" is not a block device\n", disk_name);
        goto out;
    }

    rc = fd;

out:
    return rc;
}  /* open_disk() */

static int candidate_p(const char *disk_name, int force)
{
    int fd, rc;

    rc = open_disk(disk_name);
    if (rc < 0)
        goto out;
    fd = rc;

    if (!force && !device_is_partition(fd) && device_needs_partition(fd))
    {
        rc = -EINVAL;
        if (verbose)
            error("Device \"%s\" is not a partition\n", disk_name);
        goto out_close;
    }

    rc = 0;

out_close:
    close(fd);

out:
    return rc;
}


extern char *optarg;
extern int optopt;
extern int opterr;
static int parse_options(int argc, char *argv[], char **target,
                         int *force)
{
    int c;

    opterr = 0;
    while((c = getopt(argc, argv, ":hV-:fv")) != EOF)
    {
        switch(c)
        {
            case 'h':
                usage(0, NULL);
                break;

            case 'V':
                version();
                break;

            case '-':
                if (!strcmp(optarg, "help"))
                    usage(0, NULL);
                else if (!strcmp(optarg, "version"))
                    version();
                else if (!strcmp(optarg, "force"))
                    *force = 1;
                else if (!strcmp(optarg, "verbose"))
                    verbose = 1;
                else
                    usage(-EINVAL, "Invalid option: \'--%s\'\n", optarg);
                break;

            case 'f':
                *force = 1;
                break;

            case 'v':
                verbose = 1;
                break;

            case '?':
                usage(-EINVAL, "Invalid option: \'-%c\'\n", optopt);
                break;

            default:
                usage(-EINVAL, "%s: Can't get here\n", __func__);
                break;
        }
    }

    if (optind >= argc)
        usage(-EINVAL, "You must specify a device\n");

    *target = argv[optind];
    optind++;

    /* Only one device allowed */
    if (optind < argc)
        usage(-EINVAL, "Too many arguments\n");

    return 0;
}  /* parse_options() */


/*
 * Main program
 */
int main(int argc, char *argv[])
{
    int rc, force = 0;
    char *target = NULL;

    rc = parse_options(argc, argv, &target, &force);
    if (rc)
        goto out;

    rc = candidate_p(target, force);

out:
    return rc;
}  /* main() */


