/*
 * exec_cmd.c - Execute a command in the ORACLEASM_EXEC_PATH.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 *   This code is from the GIT version control system.
 *   Copyright (c) 2006 Michal Ostrowski <mostrows@watson.ibm.com> and
 *   others.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include "usage.h"
#include "exec_cmd.h"

#define MAX_ARGS	32
#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))


extern char **environ;
static const char *builtin_exec_path = ORACLEASM_EXEC_PATH;
static const char *current_exec_path = NULL;

void as_set_exec_path(const char *exec_path)
{
	current_exec_path = exec_path;
}


/* Returns the highest-priority, location to look for as programs. */
const char *as_exec_path(void)
{
	const char *env;

	if (current_exec_path)
		return current_exec_path;

	env = getenv("ORACLEASM_EXEC_PATH");
	if (env && *env) {
		return env;
	}

	return builtin_exec_path;
}


int execv_as_cmd(const char **argv)
{
	char as_command[PATH_MAX + 1];
	int i;
	const char *paths[] = { current_exec_path,
				getenv("ORACLEASM_EXEC_PATH"),
				builtin_exec_path };

	for (i = 0; i < ARRAY_SIZE(paths); ++i) {
		size_t len;
		int rc;
		const char *exec_dir = paths[i];
		const char *tmp;

		if (!exec_dir || !*exec_dir) continue;

		if (*exec_dir != '/') {
			if (!getcwd(as_command, sizeof(as_command))) {
				error("Cannot determine "
				      "current directory: %s\n",
				      strerror(errno));
				break;
			}
			len = strlen(as_command);

			/* Trivial cleanup */
			while (!strncmp(exec_dir, "./", 2)) {
				exec_dir += 2;
				while (*exec_dir == '/')
					exec_dir++;
			}

			rc = snprintf(as_command + len,
				      sizeof(as_command) - len, "/%s",
				      exec_dir);
			if (rc < 0 || rc >= sizeof(as_command) - len) {
				error("Command name given is too long.\n");
				break;
			}
		} else {
			if (strlen(exec_dir) + 1 > sizeof(as_command)) {
				error("Command name given is too long.\n");
				break;
			}
			strcpy(as_command, exec_dir);
		}

		len = strlen(as_command);
		rc = snprintf(as_command + len, sizeof(as_command) - len,
			      "/oracleasm-%s", argv[0]);
		if (rc < 0 || rc >= sizeof(as_command) - len) {
			error("command name given is too long.\n");
			break;
		}

		/* argv[0] must be the as command, but the argv array
		 * belongs to the caller, and my be reused in
		 * subsequent loop iterations. Save argv[0] and
		 * restore it on error.
		 */

		tmp = argv[0];
		argv[0] = as_command;

		/* execve() can only ever return if it fails */
		execve(as_command, (char **)argv, environ);

		argv[0] = tmp;
	}
	return -1;

}


int execl_as_cmd(const char *cmd,...)
{
	int argc;
	const char *argv[MAX_ARGS + 1];
	const char *arg;
	va_list param;

	va_start(param, cmd);
	argv[0] = cmd;
	argc = 1;
	while (argc < MAX_ARGS) {
		arg = argv[argc++] = va_arg(param, char *);
		if (!arg)
			break;
	}
	va_end(param);
	if (MAX_ARGS <= argc) {
		error("Too many args to run %s\n", cmd);
		return -1;
	}

	argv[argc] = NULL;
	return execv_as_cmd(argv);
}
