/*
 * label.c - Read and write the label of an Oracle ASMLib disk.
 *
 * Read mode has simple semantics.  If it returns 0, it successfully
 * read the disk.  If it found a valid ASM header, it will output the
 * label, a colon (':'), then the device name and a newline to standard
 * output -- an emtpy label can still be parsed as the empty string before
 * the colon.  If it did not find a header, it will output nothing.  If
 * there is an error accessing a disk, it will return -errno and print an
 * error on standard error.
 *
 * Write mode is similar.  A return code of 0 means that a label was
 * successfully written.  If an error occurs, it will return -errno and
 * print an error on standard error.
 *
 * This program explicitly disallows multiple targets so that error
 * reporting can be explicitly assigned.
 *
 * Copyright (c) 2002-2006 Oracle Corporation.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#define _GNU_SOURCE             /*
                                 * For bits/fcntl.h O_DIRECT.  See,
                                 * ppc O_DIRECT is a differnet number,
                                 * so I can't self-define it.  And
                                 * glibc hides it under __USE_GNU,
                                 * becuase it's only on GNU...wait,
                                 * it's not.
                                 * I HATE you GLIBC.
                                 */
#define _XOPEN_SOURCE 600       /* For pread() and posix_memalign() */
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE     /* For open64() and pread64() */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

#include <linux/types.h>

#ifndef ORACLEASM_OLD_HEADERS
# include "linux/oracleasm/compat32.h"
# include "linux/oracleasm/abi.h"
# include "linux/oracleasm/disk.h"
# include "linux/oracleasm/manager.h"
#else
# include "linux/asmcompat32.h"
# include "linux/asmabi.h"
# include "linux/asmdisk.h"
# include "linux/asmmanager.h"
#endif

#include "kernel-list.h"
#include "devmap.h"
#include "usage.h"

#ifndef O_DIRECT
# define O_DIRECT 00040000  /* This is wrong for non-x86, but I hope that
                               _GNU_SOURCE picked up the correct one. */
#endif

#ifndef BLKSSZGET
# define BLKSSZGET  _IO(0x12,104)
#endif


enum LabelMode {
    NO_MODE = 0,
    READ_LABEL = 1 << 0,        /* oracleasm-read-label */
    WRITE_LABEL = 1 << 1,       /* oracleasm-write-label */
    FORCE_LABEL = 1 << 2,       /* Force changing an existing label */
    CLEAR_LABEL = 1 << 3,       /* Clear the label and set ORCLCLRD */
    CLEAR_DISK = 1 << 4,        /* Clear the head and tail of the disk
                                   (new label) */
};


/*
 * Defines
 */
#define ASM_DISK_XOR_OFFSET     0x0C
/* Ted T'so recommends clearing 16K at the front and end of disks. */
#define CLEAR_DISK_SIZE         (16 * 1024)

/*
 * Globals
 */
#define PROGNAME_READ "oracleasm-read-label"
#define PROGNAME_WRITE "oracleasm-write-label"
const char *program_name;
const char *program_read = PROGNAME_READ;
const char *program_write = PROGNAME_WRITE;

const char *program_usage;
const char *usage_read = (
    "Usage: " PROGNAME_READ " <device>\n"
    "       " PROGNAME_READ " -h\n"
    "       " PROGNAME_READ " -V\n"
);
const char *usage_write = (
    "To label: " PROGNAME_WRITE " [-f] [-c] <device> <label>\n"
    "To clear: " PROGNAME_WRITE " -d <device>\n"
    "          " PROGNAME_WRITE " -h\n"
    "          " PROGNAME_WRITE " -V\n"
);



/*
 * Typedefs
 */
typedef struct _ASMHeaderInfo   ASMHeaderInfo;



/*
 * Structures
 */
struct _ASMHeaderInfo
{
    __u32 ah_xor;
    struct asm_disk_label ah_label;
};




/*
 * Functions
 */


#ifndef __USE_GNU
/*
 * strnlen() is useful, and it's under __USE_GNU
 */
static inline size_t strnlen(const char *s, size_t size)
{
    char *p = memchr(s, '\0', size);
    return p ? p - s : size;
}  /* strnlen() */
#endif


static int open_disk(const char *disk_name, int write)
{
    int fd, rc, mode = O_DIRECT;
    struct stat64 stat_buf;

    rc = stat64(disk_name, &stat_buf);
    if (rc)
        return -errno;

    /* First check, try to avoid opening tapes */
    if (!S_ISBLK(stat_buf.st_mode))
        return -ENOTBLK;

    if (write)
        mode |= O_RDWR | O_EXCL;
    else
        mode |= O_RDONLY;
    fd = open64(disk_name, mode);
    if (fd < 0)
        return -errno;

    /* Second check, try not to write the wrong place */
    rc = fstat64(fd, &stat_buf);
    if (rc)
    {
        close(fd);
        return -errno;
    }
    
    if (!S_ISBLK(stat_buf.st_mode))
    {
        close(fd);
        return -ENOTBLK;
    }

    return fd;
}  /* open_disk() */


static int read_disk(int fd, ASMHeaderInfo *ahi)
{
    ssize_t ret;
    int rc, tot, sector_size = 0;
    void *stupid;
    char *block;

    if ((fd < 0) || !ahi)
        return -EINVAL;

    rc = ioctl(fd, BLKSSZGET, &sector_size);
    if (rc || !sector_size)  
        sector_size = 512;  /* Best try */

    rc = posix_memalign(&stupid, sector_size, sector_size);
    if (rc)
        goto out;

    block = stupid;

    tot = 0;
    while (tot < sector_size)
    {
        ret = pread64(fd, block + tot, sector_size - tot, tot);
        if (!ret)
            break;
        if (ret < 0)
        {
            if ((errno == EAGAIN) || (errno == EINTR))
                continue;
            rc = -errno;
            goto out_free;
        }

        tot += ret;
    }

    rc = -ENOSPC;
    if (tot < sector_size)
        goto out_free;

    memcpy(&ahi->ah_xor, block + ASM_DISK_XOR_OFFSET, sizeof(__u32));
    memcpy(&ahi->ah_label, block + ASM_DISK_LABEL_OFFSET,
           sizeof(struct asm_disk_label));

    rc = 0;

out_free:
    free(block);

out:
    return rc;
}  /* read_disk() */


static int write_disk(int fd, ASMHeaderInfo *ahi)
{
    ssize_t ret;
    int rc, tot, sector_size = 0;
    void *stupid;
    char *block;

    if ((fd < 0) || !ahi)
        return -EINVAL;

    rc = ioctl(fd, BLKSSZGET, &sector_size);
    if (rc || !sector_size)  
        sector_size = 512;  /* Best try */

    rc = posix_memalign(&stupid, sector_size, sector_size);
    if (rc)
        goto out;

    block = stupid;

    /* We're doing Read-Modify-Write here */
    tot = 0;
    while (tot < sector_size)
    {
        ret = pread64(fd, block + tot, sector_size - tot, tot);
        if (!ret)
            break;
        if (ret < 0)
        {
            if ((errno == EAGAIN) || (errno == EINTR))
                continue;
            rc = -errno;
            goto out_free;
        }

        tot += ret;
    }

    rc = -ENOSPC;
    if (tot < sector_size)
        goto out_free;

    memcpy(block + ASM_DISK_XOR_OFFSET, &ahi->ah_xor, sizeof(__u32));
    memcpy(block + ASM_DISK_LABEL_OFFSET, &ahi->ah_label, 
           sizeof(struct asm_disk_label));

    tot = 0;
    while (tot < sector_size)
    {
        ret = pwrite64(fd, block + tot, sector_size - tot, tot);
        if (!ret)
            break;
        if (ret < 0)
        {
            if ((errno == EAGAIN) || (errno == EINTR))
                continue;
            rc = -errno;
            goto out_free;
        }

        tot += ret;
    }

    rc = -ENOSPC;
    if (tot < sector_size)
        goto out_free;

    rc = 0;

out_free:
    free(block);

out:
    return 0;
}  /* write_disk() */


static int clear_disk_chunk(int fd, char *buf)
{
    int rc, tot = 0;

    while (tot < CLEAR_DISK_SIZE)
    {
        rc = write(fd, buf + tot, CLEAR_DISK_SIZE - tot);
        if (!rc)
            break;
        if (rc < 0)
        {
            if ((errno == EAGAIN) || (errno == EINTR))
                continue;
            else
                return -errno;
        }
        tot += rc;
    }

    if (tot < CLEAR_DISK_SIZE)
        return -ENOSPC;

    return 0;
}


static int clear_disk(int fd)
{
    loff_t off;
    int rc, sector_size;
    char *buf;
    void *stupid;

    if (fd < 0)
        return -EINVAL;

    rc = ioctl(fd, BLKSSZGET, &sector_size);
    if (rc || !sector_size)  
        sector_size = 512;  /* Best try */

    rc = posix_memalign(&stupid, sector_size, CLEAR_DISK_SIZE);
    if (rc)
        return rc;

    buf = stupid;
    memset(buf, 0, CLEAR_DISK_SIZE);

    off = lseek64(fd, 0, SEEK_SET);
    if (off == (loff_t)-1) {
        rc = -errno;
        goto out_free;
    }

    rc = clear_disk_chunk(fd, buf);
    if (rc)
        goto out_free;

    off = lseek64(fd, -CLEAR_DISK_SIZE, SEEK_END);
    if (off == (loff_t)-1) {
        rc = -errno;
        goto out_free;
    }

    rc = clear_disk_chunk(fd, buf);

out_free:
    free(buf);

    return rc;
}


static __u32 xor_buffer(const char *buf, int bufsize)
{
    int i;
    __u32 xor_buf = 0;
    __u32 *xor_ptr = (__u32 *)buf;

    for (i = 0; i < (bufsize / sizeof(__u32)); i++)
        xor_buf ^= xor_ptr[i];

    return xor_buf;
}  /* xor_buffer() */


static int mark_disk(int fd, ASMHeaderInfo *ahi, const char *label)
{
    int rc;
    size_t len;
    __u32 xor_buf;

    if ((fd < 0) || !ahi || !label || !*label)
        return -EINVAL;

    memcpy(ahi->ah_label.dl_tag, ASM_DISK_LABEL_MARKED,
           sizeof(ahi->ah_label.dl_tag));

    xor_buf = xor_buffer(ahi->ah_label.dl_id,
                         sizeof(ahi->ah_label.dl_id));

    memset(ahi->ah_label.dl_id, 0, sizeof(ahi->ah_label.dl_id));
    len = strnlen(label, sizeof(ahi->ah_label.dl_id));
    memcpy(ahi->ah_label.dl_id, label, len);

    xor_buf ^= xor_buffer(ahi->ah_label.dl_id,
                          sizeof(ahi->ah_label.dl_id));

    ahi->ah_xor ^= xor_buf;

    rc = write_disk(fd, ahi);

    return rc;
}  /* mark_disk() */


static int unmark_disk(int fd, ASMHeaderInfo *ahi)
{
    int rc;

    if ((fd < 0) || !ahi)
        return -EINVAL;

    memcpy(ahi->ah_label.dl_tag, ASM_DISK_LABEL_CLEAR,
           sizeof(ahi->ah_label.dl_tag));
    memset(ahi->ah_label.dl_id, 0, sizeof(ahi->ah_label.dl_id));

    rc = write_disk(fd, ahi);

    return rc;
}  /* unmark_disk() */


static int check_disk(ASMHeaderInfo *ahi)
{
    if (!ahi)
        return -EINVAL;

    /* Not an ASM disk */
    if (memcmp(ahi->ah_label.dl_tag, ASM_DISK_LABEL_MARKED,
               sizeof(ahi->ah_label.dl_tag)))
        return -ENXIO;

    return 0;
}  /* check_disk() */


static char *asm_disk_id(ASMHeaderInfo *ahi)
{
    char *id;

    id = (char *)malloc(sizeof(char) *
                        (sizeof(ahi->ah_label.dl_id) + 1));
    if (id)
    {
        memcpy(id, ahi->ah_label.dl_id, sizeof(ahi->ah_label.dl_id));
        id[sizeof(ahi->ah_label.dl_id)] = '\0';
    }

    return id;
}  /* asm_disk_id() */


static int get_info(const char *target)
{
    int rc, fd;
    char *id;
    ASMHeaderInfo ahi;

    rc = open_disk(target, 0);
    if (rc < 0)
    {
        error("Unable to open device \"%s\": %s\n", target, strerror(-rc));
        goto out;
    }
    fd = rc;

    rc = read_disk(fd, &ahi);
    if (rc)
    {
        error("Unable to read device \"%s\": %s\n", target, strerror(-rc));
        goto out_close;
    }

    rc = check_disk(&ahi);
    if (!rc)
    {
        id = asm_disk_id(&ahi);
        if (!id)
        {
            error("Unable to allocate memory\n");
            goto out_close;
        }

        fprintf(stdout, "%s:%s\n", id, target);
        free(id);
    }
    else if (rc == -ENXIO)
        rc = 0; /* Successful read, just not an ASMLib disk */
    else
        error("%s: Can't get here!\n", __func__);

out_close:
    close(fd);

out:
    return rc;
}  /* get_info() */


static int set_label(int fd, ASMHeaderInfo *ahi, int mode,
                     const char *target, const char *label)
{
    int rc;
    char *id;

    if (!(mode & FORCE_LABEL))
    {
        rc = check_disk(ahi);
        if (!rc)
        {
            id = asm_disk_id(ahi);
            if (!id)
                error("Unable to allocate memory\n");
            else
            {
                error("Device \"%s\" is already labeled \"%s\"\n", target,
                      id);
                free(id);
            }

            rc = -EEXIST;
            goto out;
        }
        else if (rc == -ENXIO)
            rc = 0; /* Successful read, just not an ASMLib disk */
        else
        {
            error("%s: Can't get here!\n", __func__);
            goto out;
        }
    }

    if (mode & CLEAR_DISK)
    {
        rc = clear_disk(fd);
        if (rc)
        {
            error("Unable to clear device \"%s\": %s\n", target,
                  strerror(-rc));
            goto out;
        }
    }

    rc = mark_disk(fd, ahi, label);
    if (rc)
        error("Unable to label device \"%s\": %s\n", target, strerror(-rc));

out:
    return rc;
}


static int clear_label(int fd, ASMHeaderInfo *ahi, const char *target)
{
    int rc;

    rc = check_disk(ahi);
    if (rc == -ENXIO)
        return 0;
    else if (rc)
    {
         error("%s: Can't get here!\n", __func__);
         goto out;
    }

    rc = unmark_disk(fd, ahi);
    if (rc)
        error("Unable to remove label from device \"%s\": %s\n", target,
              strerror(-rc));

out:
    return rc;
}


static int set_info(int mode, const char *target, const char *label)
{
    int rc, fd;
    ASMHeaderInfo ahi;

    rc = open_disk(target, 1);
    if (rc < 0)
    {
        error("Unable to open device \"%s\": %s\n", target, strerror(-rc));
        goto out;
    }
    fd = rc;

    rc = read_disk(fd, &ahi);
    if (rc)
    {
        error("Unable to read device \"%s\": %s\n", target, strerror(-rc));
        goto out_close;
    }

    if (mode & CLEAR_LABEL)
        rc = clear_label(fd, &ahi, target);
    else
        rc = set_label(fd, &ahi, mode, target, label);

out_close:
    close(fd);

out:
    return rc;
}


extern char *optarg;
extern int optopt;
extern int opterr;
static int parse_options(int argc, char *argv[], int *mode,
                         char **target, char **label)
{
    int c;

    opterr = 0;
    while((c = getopt(argc, argv, ":hV-:dcf")) != EOF)
    {
        switch(c)
        {
            case 'h':
                usage(0, NULL);
                break;

            case 'V':
                version();
                break;

            case '-':
                if (!strcmp(optarg, "help"))
                    usage(0, NULL);
                else if (!strcmp(optarg, "version"))
                    version();
                else
                    usage(-EINVAL, "Invalid option: \'--%s\'\n", optarg);
                break;

            case 'd':
                if (!(*mode & WRITE_LABEL))
                    usage(-EINVAL, "Invalid option: \'-d\'\n");
                *mode |= CLEAR_LABEL;
                break;

            case 'f':
                if (!(*mode & WRITE_LABEL))
                    usage(-EINVAL, "Invalid option: \'-f\'\n");
                *mode |= FORCE_LABEL;
                break;

            case 'c':
                if (!(*mode & WRITE_LABEL))
                    usage(-EINVAL, "Invalid option: \'-c\'\n");
                *mode |= CLEAR_DISK;
                break;

            case '?':
                usage(-EINVAL, "Invalid option: \'-%c\'\n", optopt);
                break;

            default:
                usage(-EINVAL, "%s: Can't get here\n", __func__);
                break;
        }
    }

    if (optind >= argc)
        usage(-EINVAL, "You must specify a device\n");

    *target = argv[optind];
    optind++;

    if ((*mode & WRITE_LABEL) && !(*mode & CLEAR_LABEL))
    {
        if (optind >= argc)
            usage(-EINVAL, "You must specify a label\n");
        *label = argv[optind];
        optind++;

        c = asmdisk_toupper(*label, -1, 0);
        if (c)
            usage(-EINVAL, "Invalid characters in label: \'%c\'\n", c);
    }

    /* Only one device allowed */
    if (optind < argc)
        usage(-EINVAL, "Too many arguments\n");

    return 0;
}  /* parse_options() */


/*
 * Main program
 */
int main(int argc, char *argv[])
{
    char *cmd = argv[0];
    char *slash = strrchr(cmd, '/');
    int rc, mode = NO_MODE;
    char *target = NULL;
    char *label = NULL;

    if (slash)
    {
        *slash++ = 0;
        cmd = slash;
    }

    program_name = cmd;
    argv[0] = cmd;

    if (!strcmp(cmd, program_read))
    {
        mode = READ_LABEL;
        program_usage = usage_read;
    }
    else if (!strcmp(cmd, program_write))
    {
        mode = WRITE_LABEL;
        program_usage = usage_write;
    }
    else
    {
        rc = -ENOENT;
        program_name = cmd;
        error("label.c cannot handle \"%s\" internally\n", cmd);
        goto out;
    }

    rc = parse_options(argc, argv, &mode, &target, &label);
    if (rc)
        goto out;

    if (mode & READ_LABEL)
        rc = get_info(target);
    else if (mode & WRITE_LABEL)
        rc = set_info(mode, target, label);
    else
    {
        rc = -EINVAL;
        error("%s: Can't get here!\n", __func__);
    }

out:
    return rc;
}  /* main() */


