/* -*- mode: c; c-basic-offset: 8; -*-
 * vim: noexpandtab sw=8 ts=8 sts=0:
 *
 * exec_cmd.c - Execute a command in the ORACLEASM_EXEC_PATH.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 *   This code is from the GIT version control system, specifically the
 *   git.c wrapper, Copyright (c) 2005 Andreas Ericsson <exon@op5.se> and
 *   others, and builtin-help.c, Copyright (c) 2005 Linus Torvalds
 *   <torvalds@osdl.org>.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <ctype.h>
#include <stdarg.h>

#include "usage.h"
#include "exec_cmd.h"

#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))

#define PROGNAME "oracleasm"
const char *program_name = PROGNAME;
const char *program_usage = (
	"Usage: " PROGNAME " [--exec-path=<exec_path>] <command> [ <args> ]\n"
	"       " PROGNAME " --exec-path\n"
	"       " PROGNAME " -h\n"
	"       " PROGNAME " -V\n"
);

struct cmdname_help
{
	char name[32];
	char help[64];
};

static struct cmdname_help normal_commands[] = {
	{"configure", "Configure the Oracle Linux ASMLib driver"},
	{"init", "Load and initialize the ASMLib driver"},
	{"exit", "Stop the ASMLib driver"},
	{"scandisks", "Scan the system for Oracle ASMLib disks"},
	{"status", "Display the status of the Oracle ASMLib driver"},
	{"listdisks", "List known Oracle ASMLib disks"},
	{"querydisk", "Determine if a disk belongs to Oracle ASMlib"},
	{"createdisk", "Allocate a device for Oracle ASMLib use"},
	{"deletedisk", "Return a device to the operating system"},
	{"renamedisk", "Change the label of an Oracle ASMlib disk"},
	{"update-driver", "Download the latest ASMLib driver"},
};


/* most GUI terminals set COLUMNS (although some don't export it) */
static int term_columns(void)
{
	char *col_string = getenv("COLUMNS");
	int n_cols = 0;

	if (col_string && (n_cols = atoi(col_string)) > 0)
		return n_cols;

#ifdef TIOCGWINSZ
	{
		struct winsize ws;
		if (!ioctl(1, TIOCGWINSZ, &ws)) {
			if (ws.ws_col)
				return ws.ws_col;
		}
	}
#endif

	return 80;
}

static inline void mput_char(char c, unsigned int num)
{
	while(num--)
		putchar(c);
}

static void oom(void)
{
	error("Out of memory\n");
	exit(1);
}

static struct cmdname {
	size_t len;
	char name[1];
} **cmdname;
static int cmdname_alloc, cmdname_cnt;

static void add_cmdname(const char *name, int len)
{
	struct cmdname *ent;
	if (cmdname_alloc <= cmdname_cnt) {
		cmdname_alloc = cmdname_alloc + 200;
		cmdname = realloc(cmdname, cmdname_alloc * sizeof(*cmdname));
		if (!cmdname)
			oom();
	}
	ent = malloc(sizeof(*ent) + len);
	if (!ent)
		oom();
	ent->len = len;
	memcpy(ent->name, name, len);
	ent->name[len] = 0;
	cmdname[cmdname_cnt++] = ent;
}

static int cmdname_compare(const void *a_, const void *b_)
{
	struct cmdname *a = *(struct cmdname **)a_;
	struct cmdname *b = *(struct cmdname **)b_;
	return strcmp(a->name, b->name);
}

static void pretty_print_string_list(struct cmdname **cmdname, int longest)
{
	int cols = 1, rows;
	int space = longest + 1; /* min 1 SP between words */
	int max_cols = term_columns() - 1; /* don't print *on* the edge */
	int i, j;

	if (space < max_cols)
		cols = max_cols / space;
	rows = (cmdname_cnt + cols - 1) / cols;

	qsort(cmdname, cmdname_cnt, sizeof(*cmdname), cmdname_compare);

	for (i = 0; i < rows; i++) {
		printf("  ");

		for (j = 0; j < cols; j++) {
			int n = j * rows + i;
			int size = space;
			if (n >= cmdname_cnt)
				break;
			if (j == cols-1 || n + rows >= cmdname_cnt)
				size = 1;
			printf("%-*s", size, cmdname[n]->name);
		}
		putchar('\n');
	}
}

static void list_commands(const char *exec_path)
{
	unsigned int longest = 0;
	unsigned int prelen = strlen("oracleasm-");
	char path[PATH_MAX];
	int dirlen;
	DIR *dir = opendir(exec_path);
	struct dirent *de;

	if (!dir) {
		error("'%s': %s\n", exec_path, strerror(errno));
		exit(1);
	}

	dirlen = strlen(exec_path);
	if (PATH_MAX - 20 < dirlen) {
		error("Insanely long exec-path '%s'\n", exec_path);
		exit(1);
	}

	memcpy(path, exec_path, dirlen);
	path[dirlen++] = '/';

	while ((de = readdir(dir)) != NULL) {
		struct stat st;
		int entlen;

		if (strncmp(de->d_name, "oracleasm-", prelen))
			continue;
		/* Skip script libraries */
		if (de->d_name[prelen] == 'X')
			continue;
		strcpy(path+dirlen, de->d_name);
		if (stat(path, &st) || /* stat, not lstat */
		    !S_ISREG(st.st_mode) ||
		    !(st.st_mode & S_IXUSR))
			continue;

		entlen = strlen(de->d_name);
		if (4 < entlen && !strcmp(de->d_name + entlen - 4, ".exe"))
			entlen -= 4;

		if (longest < entlen)
			longest = entlen;

		add_cmdname(de->d_name + prelen, entlen - prelen);
	}
	closedir(dir);

	fprintf(stdout, "oracleasm commands available in '%s'\n",
		exec_path);
	fprintf(stdout, "----------------------------");
	mput_char('-', strlen(exec_path));
	putchar('\n');
	pretty_print_string_list(cmdname, longest - 4);
	putchar('\n');
}

static void list_normal_commands(void)
{
	int i, longest = 0;

	for (i = 0; i < ARRAY_SIZE(normal_commands); i++) {
		if (longest < strlen(normal_commands[i].name))
			longest = strlen(normal_commands[i].name);
	}

	fprintf(stdout, "The basic oracleasm commands are:\n");
	for (i = 0; i < ARRAY_SIZE(normal_commands); i++) {
		fprintf(stdout, "    %s", normal_commands[i].name);
		mput_char(' ', longest - strlen(normal_commands[i].name) + 4);
		fprintf(stdout, "%s\n", normal_commands[i].help);
	}
}

static void core_usage(int rc, int show_all, const char *exec_path, const char *fmt, ...)
{
	FILE *output = rc ? stderr : stdout;

	if (fmt) {
		va_list ap;

		va_start(ap, fmt);
		fprintf(output, "%s: ", program_name);
		vfprintf(output, fmt, ap);
		va_end(ap);
	}

	fprintf(output, "%s", program_usage);

	if (exec_path) {
		fputc('\n', output);

		if (show_all)
			list_commands(exec_path);
		else
			list_normal_commands();
	}

	exit(rc);
}


static void core_cmd(const char **argv)
{
	execv_as_cmd(argv);

	if (errno == ENOENT)
		core_usage(-ENOENT, 0, as_exec_path(),
			   "'%s' is not an oracleasm command\n",
			   argv[0]);

	error("Failed to run command '%s': %s\n", argv[0], strerror(errno));
}


static void core_help(int argc, const char **argv, char **env)
{
	const char *help_cmd = argv[1];
	const char *help_argv[4];

	if (!help_cmd)
		core_usage(0, 0, as_exec_path(), NULL);
	else if (!strcmp(help_cmd, "--all") || !strcmp(help_cmd, "-a"))
		core_usage(0, 1, as_exec_path(), NULL);
	else {
		help_argv[0] = help_argv[1] = help_cmd;
		help_argv[2] = "--help";
		help_argv[3] = NULL;

		core_cmd(help_argv);
	}

	exit(0);
}


static void prepend_to_path(const char *dir, int len)
{
	const char *old_path = getenv("PATH");
	char *path;
	int path_len = len;

	if (!old_path)
		old_path = "/usr/local/bin:/usr/bin:/bin";

	path_len = len + strlen(old_path) + 1;

	path = malloc(path_len + 1);

	memcpy(path, dir, len);
	path[len] = ':';
	memcpy(path + len + 1, old_path, path_len - len);

	setenv("PATH", path, 1);
}


int main(int argc, const char **argv, char **envp)
{
	const char *cmd = argv[0];
	char *slash = strrchr(cmd, '/');
	const char *exec_path = NULL;

	/*
	 * Take the basename of argv[0] as the command
	 * name, and the dirname as the default exec_path
	 * if it's an absolute path and we don't have
	 * anything better.
	 */
	if (slash) {
		*slash++ = 0;
		if (*cmd == '/')
			exec_path = cmd;
		cmd = slash;
	}

	/* Default command: "help" */
	cmd = "help";

	/* Look for flags.. */
	while (argc > 1) {
		cmd = *++argv;
		argc--;

		if (strncmp(cmd, "--", 2))
			break;

		cmd += 2;

		/*
		 * For legacy reasons, the "version" and "help"
		 * commands can be written with "--" prepended
		 * to make them look like flags.
		 */
		if (!strcmp(cmd, "h"))
			cmd = "help";
		if (!strcmp(cmd, "V"))
			cmd = "version";
		if (!strcmp(cmd, "help"))
			break;
		if (!strcmp(cmd, "version"))
			break;

		/*
		 * Check remaining flags (which by now must be
		 * "--exec-path", but maybe we will accept
		 * other arguments some day)
		 */
		if (!strncmp(cmd, "exec-path", 9)) {
			cmd += 9;
			if (*cmd == '=') {
				as_set_exec_path(cmd + 1);
				continue;
			}
			puts(as_exec_path());
			exit(0);
		}
		core_usage(-EINVAL, 0, exec_path,
			   "Invalid option: \'--%s\'\n", cmd);
	}
	argv[0] = cmd;

	/*
	 * We search for oracleasm commands in the following order:
	 *  - as_exec_path()
	 *  - the path of the "oracleasm" command if we could find it
	 *    in $0
	 *  - the regular PATH.
	 *
	 *  Note: the source in exec_cmd no longer seems to do this,
	 *  even in GIT.  But programs using PATH (eg, the shell ones)
	 *  will.
	 */
	if (exec_path)
		prepend_to_path(exec_path, strlen(exec_path));
	exec_path = as_exec_path();
	prepend_to_path(exec_path, strlen(exec_path));

	if (!strcmp(cmd, "help"))
		core_help(argc, argv, envp);
	if (!strcmp(cmd, "version"))
		version();

	core_cmd(argv);

	return 1;
}
