/*
 * reload-partitions.c - Scan system for block devices and reload
 *                       their partition tables.
 *
 * Copyright (c) 2002-2006 Oracle Corporation.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include <linux/types.h>
#include <linux/hdreg.h>

#include "usage.h"


/*
 * Defines
 */

/* So as not to include linux/fs.h, let's explicitly do this here. */
#ifndef BLKRRPART
#define BLKRRPART  _IO(0x12,95) /* re-read partition table */
#endif


/*
 * Globals
 */
#define PROGNAME "oracleasm-reload-partitions"
const char *program_name = PROGNAME;
const char *program_usage = (
            "Usage: " PROGNAME " < device_list\n"
            "       " PROGNAME " -h\n"
            "       " PROGNAME " -V\n"
);
 


/*
 * Functions
 */


static int device_is_partition(const char *device)
{
    struct hd_geometry geo;
    int rc, fd;

    /* Any errors, return false.  This will force BLKRRPART on the device. */
    fd = open64(device, O_RDONLY);
    if (fd < 0)
        return 0;

    rc = ioctl(fd, HDIO_GETGEO, &geo);

    close(fd);

    if (rc)
        return 0;

    return !!geo.start;
}  /* device_is_partition() */


static int reread_single_device(const char *device)
{
    int rc, fd;

    /* Any errors, return false.  This will force BLKRRPART on the device. */
    fd = open64(device, O_RDONLY);
    if (fd < 0)
        return -errno;

    rc = ioctl(fd, BLKRRPART, NULL);

    if (rc)
        rc = -errno;

    close(fd);

    return rc;
}  /* reread_single_device() */


static int reread_partitions(void)
{
    int rc, len;
    char device[PATH_MAX];

    while (fgets(device, PATH_MAX, stdin) != NULL)
    {
        len = strlen(device);
        if (device[len - 1] == '\n')
            device[len - 1] = '\0';

        if (device_is_partition(device))
            continue;
        rc = reread_single_device(device);
        /*
         * Drop rc to the fllor, as failing to reread partitions
         * just means we go with the existing partition map.
         * (Keep rc as a variable, just in case we care later.)
         */
    }

    /* Ignore errors */
    return 0;
}  /* rescan_partitions() */


extern char *optarg;
extern int optopt;
extern int opterr;
static int parse_options(int argc, char *argv[])
{
    int c;

    opterr = 0;
    while((c = getopt(argc, argv, ":-:hV")) != EOF)
    {
        switch(c)
        {
            case 'h':
                usage(0, NULL);
                break;

            case 'V':
                version();
                break;

            case '-':
                if (!strcmp(optarg, "help"))
                    usage(0, NULL);
                else if (!strcmp(optarg, "version"))
                    version();
                else
                    usage(-EINVAL, "Invalid option: \'--%s\'\n", optarg);
                break;

            case '?':
                usage(-EINVAL, "Invalid option: \'-%c\'\n", optopt);
                break;

            case ':':
                usage(-EINVAL, "Option \'-%c\' requires an argument\n",
                      optopt);
                break;

            default:
                usage(-EINVAL, "Can't get here!\n");
                break;
        }
    }

    if (optind < argc)
        usage(-EINVAL, "Too many arguments\n");

    return 0;
}  /* parse_options() */



/*
 * Main program
 */
int main(int argc, char *argv[])
{
    int rc;

    rc = parse_options(argc, argv);
    if (!rc)
        rc = reread_partitions();

    return rc;
}  /* main() */


