/*
 * NAME
 *	partitions.c - Scan /proc/partitions for a list of known devices.
 *
 * AUTHOR
 * 	Joel Becker <joel.becker@oracle.com>
 *
 * Copyright (c) 2006 Oracle Corporation.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License, version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <getopt.h>
#include <errno.h>

#include <linux/types.h>

#include "kernel-list.h"
#include "usage.h"


/*
 * Defines
 */
#define DEV_PREFIX      "/dev/"
#define PROC_IDE_FORMAT "/proc/ide/%s/media"


/*
 * Typedefs
 */
typedef struct _ASMScanPattern  ASMScanPattern;
typedef struct _ASMScanDevice   ASMScanDevice;


/*
 * Structures
 */
struct _ASMScanPattern
{
    struct list_head sp_list;
    struct list_head sp_matches;
    const char *sp_pattern;
    size_t sp_len;
};

struct _ASMScanDevice
{
    struct list_head sd_list;
    char *sd_name;
    char *sd_path;
};


/*
 * Globals
 */
#define PROGNAME "oracleasm-scan-partitions"
const char *program_name = PROGNAME;
const char *program_usage = (
    "Usage: " PROGNAME " [-o <pattern> ...] [-x <pattern>]\n"
    "       " PROGNAME " -h\n"
    "       " PROGNAME " -V\n"
);


/*
 * Functions
 */


/*
 * strnlen() is useful, and it's under __USE_GNU
 */
static inline size_t strnlen(const char *s, size_t size)
{
    char *p = memchr(s, '\0', size);
    return p ? p - s : size;
}  /* strnlen() */


static ASMScanPattern *as_pattern_new(struct list_head *pattern_list,
                                      const char *text)
{
    ASMScanPattern *pattern;

    if (!text)
        return NULL;

    pattern = (ASMScanPattern *)malloc(sizeof(ASMScanPattern));
    if (!pattern)
        return NULL;

    INIT_LIST_HEAD(&pattern->sp_matches);

    pattern->sp_pattern = text;
    pattern->sp_len = strnlen(text, PATH_MAX);

    list_add_tail(&pattern->sp_list, pattern_list);

    return pattern;
}  /* as_pattern_new() */


static void as_device_free(ASMScanDevice *device)
{
    if (device)
    {
        list_del(&device->sd_list);
        if (device->sd_name)
            free(device->sd_name);
        if (device->sd_path)
            free(device->sd_path);
        free(device);
    }
}  /* as_device_free() */


static void as_pattern_empty(ASMScanPattern *pattern)
{
    struct list_head *pos, *tmp;
    ASMScanDevice *device;

    if (pattern)
    {
        if (!list_empty(&pattern->sp_matches))
        {
            list_for_each_safe(pos, tmp, &pattern->sp_matches) {
                device = list_entry(pos, ASMScanDevice, sd_list);
                as_device_free(device);
            }
        }
    }
}  /* as_pattern_empty() */


static void as_pattern_free(ASMScanPattern *pattern)
{
    if (pattern)
    {
        as_pattern_empty(pattern);
        list_del(&pattern->sp_list);
        free(pattern);
    }
}  /* as_pattern_free() */


static int as_pattern_match(ASMScanPattern *pattern,
                            const char *dev_name)
{
    if (!pattern || !pattern->sp_pattern || !dev_name)
        return -EINVAL;
    
    if (strnlen(dev_name, pattern->sp_len + 1) < pattern->sp_len)
        return 0;
    if (strncmp(dev_name, pattern->sp_pattern, pattern->sp_len))
        return 0;

    return 1;
}  /* as_pattern_match() */


static void as_clean_pattern_list(struct list_head *pattern_list)
{
    struct list_head *pos, *tmp;
    ASMScanPattern *pattern;

    list_for_each_safe(pos, tmp, pattern_list) {
        pattern = list_entry(pos, ASMScanPattern, sp_list);

        /* Will list_del() */
        as_pattern_free(pattern);
    }
}  /* as_clean_pattern_list() */


static ASMScanDevice *as_device_new(struct list_head *device_list,
                                    const char *dev_name)
{
    int len;
    ASMScanDevice *device;

    device = (ASMScanDevice *)malloc(sizeof(ASMScanDevice));
    if (!device)
        return NULL;

    device->sd_name = strdup(dev_name);
    len = strlen(DEV_PREFIX) + strlen(device->sd_name) + 1;
    device->sd_path = (char *)malloc(sizeof(char) * len);
    if (!device->sd_path)
    {
        free(device);
        return NULL;
    }

    snprintf(device->sd_path, len, DEV_PREFIX "%s", device->sd_name);

    list_add_tail(&device->sd_list, device_list);

    return device;
}  /* as_device_new() */


static int as_exclude_device(struct list_head *exclude_list,
                             const char *dev_name)
{
    struct list_head *pos;
    ASMScanPattern *pattern;

    list_for_each(pos, exclude_list) {
        pattern = list_entry(pos, ASMScanPattern, sp_list);
        if (as_pattern_match(pattern, dev_name))
        {
            if (!as_device_new(&pattern->sp_matches, dev_name))
                error("Error processing device \"%s\"\n", dev_name);
            return 1;
        }
    }

    return 0;
}  /* as_exclude_device() */


static int as_order_device(struct list_head *order_list,
                           const char *dev_name)
{
    struct list_head *pos;
    ASMScanPattern *pattern;

    list_for_each(pos, order_list) {
        pattern = list_entry(pos, ASMScanPattern, sp_list);
        if (as_pattern_match(pattern, dev_name))
        {
            if (!as_device_new(&pattern->sp_matches, dev_name))
                error("Error processing device \"%s\"\n", dev_name);
            return 1;
        }
    }

    return 0;
}  /* as_order_device() */


static int as_ide_disk(const char *dev_name)
{
    FILE *f;
    int is_disk = 1;
    size_t len;
    char *proc_name;

    len = strlen(PROC_IDE_FORMAT) + strlen(dev_name);
    proc_name = (char *)malloc(sizeof(char) * len);
    if (!proc_name)
        return 0;

    snprintf(proc_name, len, PROC_IDE_FORMAT, dev_name);
    
    /* If not ide, file won't exist */
    f = fopen64(proc_name, "r");
    if (f)
    {
        if (fgets(proc_name, len, f))
        {
            /* IDE devices we don't want to probe */
            if (!strncmp(proc_name, "cdrom", strlen("cdrom")) ||
                !strncmp(proc_name, "tape", strlen("tape")))
                is_disk = 0;
        }
        fclose(f);
    }

    free(proc_name);

    return is_disk;
}  /* as_ide_disk() */


/* Um, wow, this is, like, one big hardcode */
static int as_scan_devices(struct list_head *order_list,
                           struct list_head *exclude_list)
{
    int rc, major, minor;
    FILE *f;
    char *buffer, *name;

    buffer = (char *)malloc(sizeof(char) * (PATH_MAX + 1));
    if (!buffer)
        return -ENOMEM;

    name = (char *)malloc(sizeof(char) * (PATH_MAX + 1));
    if (!name)
    {
        free(buffer);
        return -ENOMEM;
    }

    f = fopen64("/proc/partitions", "r");
    if (!f)
    {
        rc = -errno;
        goto out_free;
    }

    while (1)
    {
        rc = 0;
        if ((fgets(buffer, PATH_MAX + 1, f)) == NULL)
            break;

        name[0] = '\0';
        major = minor = 0;

        /* FIXME: If this is bad, send patches */
        if (sscanf(buffer, "%d %d %*d %99[^ \t\n]",
                   &major, &minor, name) < 3)
            continue;

        if (*name && major)
        {
            if (!as_ide_disk(name))
                continue;

            if (as_exclude_device(exclude_list, name))
                continue;

            if (!as_order_device(order_list, name))
                error("Unable to order device \"%s\"!\n", name);
        }
    }

    fclose(f);

out_free:
    free(buffer);
    free(name);

    return rc;
}  /* scan_devices() */


static void output_partitions(struct list_head *order_list)
{
    struct list_head *pos_l, *pos_d;
    ASMScanPattern *pattern;
    ASMScanDevice *device;

    list_for_each(pos_l, order_list) {
        pattern = list_entry(pos_l, ASMScanPattern, sp_list);
        list_for_each(pos_d, &pattern->sp_matches) {
            device = list_entry(pos_d, ASMScanDevice, sd_list);
            fprintf(stdout, "%s\n", device->sd_path);
        }
    }
}  /* make_disks() */


extern char *optarg;
extern int optopt;
extern int opterr;
static int parse_options(int argc, char *argv[], char **manager,
                         struct list_head *order_list,
                         struct list_head *exclude_list)
{
    int c;

    opterr = 0;
    while((c = getopt(argc, argv, ":l:o:x:-:hV")) != EOF)
    {
        switch(c)
        {
            case 'h':
                usage(0, NULL);
                break;

            case 'V':
                version();
                break;

            case '-':
                if (!strcmp(optarg, "help"))
                    usage(0, NULL);
                else if (!strcmp(optarg, "version"))
                    version();
                else
                    usage(-EINVAL, "Invalid option: \'--%s\'\n", optarg);
                break;

            case 'o':
                if (!optarg || !*optarg)
                    usage(-EINVAL, "Empty order pattern\n");

                if (!as_pattern_new(order_list, optarg))
                {
                    error("Unable to allocate order pattern \"%s\"\n",
                          optarg);
                    return -ENOMEM;
                }
                break;

            case 'x':
                if (!optarg || !*optarg)
                    usage(-EINVAL, "Empty exclude pattern\n");

                if (!as_pattern_new(exclude_list, optarg))
                {
                    error("Unable to allocate exclude pattern \"%s\"\n",
                          optarg);
                    return -ENOMEM;
                }
                break;

            case '?':
                usage(-EINVAL, "Invalid option: \'-%c\'\n", optopt);
                break;

            case ':':
                usage(-EINVAL, "Option \'-%c\' requires an argument\n",
                      optopt);
                break;

            default:
                usage(-EINVAL, "Can't get here!\n");
                break;
        }
    }

    return 0;
}  /* parse_options() */



/*
 * Main program
 */
int main(int argc, char *argv[])
{
    int rc;
    char *manager = NULL;
    struct list_head order_list, exclude_list;

    INIT_LIST_HEAD(&order_list);
    INIT_LIST_HEAD(&exclude_list);

    rc = parse_options(argc, argv, &manager,
                       &order_list, &exclude_list);
    if (rc)
        goto out;

    rc = -ENOMEM;
    if (!as_pattern_new(&order_list, ""))
    {
        error("Unable to create the default pattern\n");
        goto out;
    }

    rc = as_scan_devices(&order_list, &exclude_list);
    if (rc)
        goto out;

    output_partitions(&order_list);

out:
    as_clean_pattern_list(&order_list);
    as_clean_pattern_list(&exclude_list);

    return rc;
}  /* main() */


